import os
import json
import shutil
from urllib.error import HTTPError
import urllib.request
from urllib.parse import urljoin
from typing import Any, Callable, Literal, overload

from .utils import escape_name, absjoin, version_key, PackageType, PackageInfo, PackageMetaData
from .consts import DEFAULT_REGISTRY_PATH, PACKAGE_INDEX_PATH, ROOT_DIR, SYSTEM_IDENTIFIER_DICT, CHUNK_SIZE, CACHE_PATH
from .config import registries, _default_package_type, get_cache_id
from .error import *
if not os.path.exists(_tmp:=absjoin(ROOT_DIR,"tmp/package_installation")):
	os.makedirs(_tmp)
if not os.path.exists(CACHE_PATH):
	os.makedirs(CACHE_PATH)
def download(url:str,progress_start:Callable[[int],Any]=lambda _:0,progress_step:Callable[[int],Any]=lambda _:0,progress_end:Callable[[],Any]=lambda:0):
	"""
	ファイルをダウンロードする。

	Params
	---
	url: ダウンロードするURL

	progress_start: ファイルのダウンロードの開始時に呼ばれる関数。戻り値は無視され、引数は次のとおり: size(位置引数で、ファイルの大きさ(byte単位))

	progress_step: ファイルのダウンロードが進んだ時に呼ばれる関数。戻り値は無視され、引数は以下のとおり: step(位置引数で、新たにダウンロードされたデータの大きさ(byte単位))

	progress_end: ファイルのダウンロードの終了時に呼ばれる関数。戻り値は無視され、引数はなし。

	Return
	---
	ダウンロードしたファイルのパス
	"""
	cache_path_=absjoin(CACHE_PATH,str(get_cache_id()),"urlopen")
	if not os.path.exists(cache_path_):
		os.makedirs(cache_path_)
	escaped_url=escape_name(url)
	cache_path=absjoin(cache_path_,escaped_url)
	if not os.path.exists(cache_path):
		with urllib.request.urlopen(url+("&"if "?" in url else "?")+"__ubazame_iwashi_random__="+str(get_cache_id())) as src:
			progress_start(int(src.headers["content-length"]))
			try:
				with open(cache_path,"wb") as dst:
					while tmp__:=src.read(CHUNK_SIZE):
						dst.write(tmp__)
						progress_step(len(tmp__))
			finally:
				progress_end()
	return cache_path
def download_and_open(url:str,progress_start:Callable[[int],Any]=lambda _:0,progress_step:Callable[[int],Any]=lambda _:0,progress_end:Callable[[],Any]=lambda:0):
	"""
	urllib.request.urlretの代わり。

	Params
	---
	url: ダウンロードするURL

	progress_start: ファイルのダウンロードの開始時に呼ばれる関数。戻り値は無視され、引数は次のとおり: size(位置引数で、ファイルの大きさ(byte単位))

	progress_step: ファイルのダウンロードが進んだ時に呼ばれる関数。戻り値は無視され、引数は以下のとおり: step(位置引数で、新たにダウンロードされたデータの大きさ(byte単位))

	progress_end: ファイルのダウンロードの終了時に呼ばれる関数。戻り値は無視され、引数はなし。

	Return
	---
	ダウンロードしたファイルをopen(...,"rb")で開いたもの。
	"""
	return open(download(url,progress_start,progress_step,progress_end),"rb")
def resolve_package_data(package_name:str)->tuple[Literal["git"],str]:
	"""
	パッケージの名前を解決する。

	Params
	---
	package_name: パッケージの名前

	Return
	---
	先頭がそのパッケージの種類のタプル。そのあとに続くものは、パッケージの種類によって異なり、次の通り:
	* "git": ("git",リポジトリのパス)
	"""
	names=package_name.split(":")
	reg_url:str=DEFAULT_REGISTRY_PATH
	if names[0]=="raw":
		if names[1]!="git":
			raise UnknownPackageTypeError(names[1])
		tmp=":".join(names[2:])
		repo_path=tmp
		true_name=os.path.basename(tmp)
	else:
		try:
			if names[0][0]=="@":
				reg_url=registries[names[0][1:]]

				names=names[1:]
			with download_and_open(reg_url) as res:
				tmp=json.load(res)
			for name_space in names[:-1]:
				with download_and_open(tmp["childs"][name_space]) as res:
					tmp=json.load(res)
			true_name=names[-1]
			if tmp["packages"][true_name]["type"]!="git":
				raise UnknownPackageTypeError(tmp["packages"][true_name]["type"])
			repo_path=tmp["packages"][true_name]["url"]
		except KeyError as e:
			raise PackageNotFoundError(package_name)
	if repo_path[-1]!="/":repo_path+="/"
	return("git",repo_path)
def uninstall(package_name:str)->None:
	"""
	パッケージをアンインストールする。
	
	Params
	---
	package_name: アンインストールするパッケージの名前
	"""
	if package_name in ["maiwashi"]:
		raise SelfUninstallationError()
	escaped_name=escape_name(package_name)
	if package_name not in index:
		raise PackageNotInstalledError(package_name)
	shutil.rmtree(absjoin(ROOT_DIR,"packages",escaped_name))
	del index[package_name]
	save_index()
def install(package_name:str,package_info:PackageInfo|None,is_update:bool,progress_start:Callable[[int],Any]=lambda _:0,progress_step:Callable[[int],Any]=lambda _:0,progress_end:Callable[[],Any]=lambda:0)->bool:
	"""
	パッケージをインストール(or アップデート)する。
	アップデートとは銘打っているが、ダウングレードも可能。

	Params
	---
	package_name: インストール(or アップデート)するパッケージの名前。

	info: インストール(or アップデート)するパッケージのパッケージ情報。

	is_update: アップデートをするかどうか(Falseなら、インストール)。

	progress_start: ファイルのダウンロードの開始時に呼ばれる関数。戻り値は無視され、引数は次のとおり: size(位置引数で、ファイルの大きさ(byte単位))

	progress_step: ファイルのダウンロードが進んだ時に呼ばれる関数。戻り値は無視され、引数は以下のとおり: step(位置引数で、新たにダウンロードされたデータの大きさ(byte単位))

	progress_end: ファイルのダウンロードの終了時に呼ばれる関数。戻り値は無視され、引数はなし。

	Return
	---
	インストール(or アップデート)を実際に行ったかどうか(アップデートで最新の状態だった場合や変更先のバージョンと今のバージョンが同じだった場合はFalse、それ以外はTrue)。
	"""
	if use_package_info:=package_info is not None:
		version=package_info.version
		package_type=package_info.package_type
		if version not in list_versions(package_name):
			raise PackageInfoNotFoundError(package_info,package_name)
		src_version=version
	else:
		version=None
		package_type=PackageType.auto
		try:
			src_version=sorted(list_versions(package_name),key=lambda x:[tmp[i]for tmp in [version_key(x)]for i in [3,0,1,2,4]])[-1]
		except IndexError:
			raise InvaidPackageError(package_name)
	_,repo_path=resolve_package_data(package_name)
	use_package_info=version is not None
	escaped_name=escape_name(package_name)
	if is_update:
		if package_name not in index:
			raise PackageNotInstalledError(package_name,in_update=True)
		_tmp=version_key(index[package_name].version)
		_tmp_=version_key(src_version)
		if  _tmp==_tmp_ or _tmp>_tmp_ and not use_package_info:
			return False
	else:
		if package_name in index:
			raise PackageAlreadyInstalledError(package_name,in_install=True)
	tmp_folder_path=absjoin(ROOT_DIR,"tmp/package_installation",escaped_name)
	folder_path=absjoin(ROOT_DIR,"packages",escaped_name)
	system_candidates:list[PackageType]
	if package_type==PackageType.auto:
		system_candidates=list(SYSTEM_IDENTIFIER_DICT)
		system_candidates.remove(_default_package_type)
		system_candidates.insert(0,_default_package_type)
	else:
		system_candidates=[package_type]
	for system in system_candidates:
		try:
			zip_path=download(urljoin(repo_path,f"releases/download/{src_version}/default-{SYSTEM_IDENTIFIER_DICT[system]}.zip"),progress_start,progress_step,progress_end)
		except HTTPError as e:
			if e.status==404:continue
			raise 
		else:
			package_type=system
			break
	else:
		if package_type!=PackageType.auto:assert package_info is not None;raise PackageInfoNotFoundError(package_info,package_name)
		raise InvaidPackageError(package_name)
	if is_update:tmp_folder_path=shutil.move(folder_path,tmp_folder_path)
	try:shutil.unpack_archive(zip_path,folder_path)
	except Exception:
		if is_update:shutil.move(tmp_folder_path,absjoin(ROOT_DIR,"packages"))
	else:
		if is_update:shutil.rmtree(tmp_folder_path)
	index[escaped_name]=PackageInfo(src_version,package_type)
	save_index()
	return True
@overload
def list_packages(include_registry:Literal[False]=False)->dict[str,PackageInfo]:...
@overload
def list_packages(include_registry:Literal[True]=True)->dict[str,PackageInfo|None]:...
def list_packages(include_registry:bool=False):
	"""
	パッケージの一覧を表示する。

	Params
	---
	include_registry: レジストリ上の、現在インストールされていないパッケージも含むかどうか。

	Return
	---
	パッケージの名前と、パッケージ情報の辞書。
	"""
	if not include_registry:return dict(index)
	index_:dict[str,PackageInfo|None]=dict(index)
	def list_package_of_registry(pref:str,url:str):
		with download_and_open(url) as res:
			tmp=json.load(res)
		for p_name in tmp["packages"]:
			true_name=pref+p_name
			if true_name not in index_:
				index_[true_name]=None
		for c_name,url in tmp["childs"].items():
			list_package_of_registry(pref+c_name+":",url)
	for name,url in registries.items():
		list_package_of_registry("@"+name+":",url)
	list_package_of_registry("",DEFAULT_REGISTRY_PATH)
	return index_
def list_versions(package_name:str)->list[str]:
	"""
	パッケージのバージョンの一覧を取得する。

	Params
	---
	package_name: バージョンの一覧を取得するパッケージの名前。

	Return
	---
	バージョンのリスト。
	"""
	_,repo_path=resolve_package_data(package_name)
	url_path=urljoin(repo_path,"raw/iwashi_package/versions/all.txt")
	with download_and_open(url_path) as res:
		tmp_=res.read().decode().splitlines()
	return tmp_
def get_meta(package_name:str)->PackageMetaData:
	"""
	パッケージのメタデータを取得する。

	Params
	---
	package_name: メタデータを取得するパッケージの名前。

	Return
	---
	取得されたメタデータ。
	"""
	_,repo_path=resolve_package_data(package_name)
	with download_and_open(urljoin(repo_path,"raw/iwashi_package/meta.json")) as src:
		return PackageMetaData(**json.load(src))
def get_readme(package_name:str)->str|None:
	"""
	パッケージの紹介文を取得する。

	Params
	---
	package_name: 紹介文を取得するパッケージの名前。

	Return
	---
	取得された紹介文。なければNone
	"""
	_,repo_path=resolve_package_data(package_name)
	try:
		with download_and_open(urljoin(repo_path,"raw/iwashi_package/readme.txt")) as src:
			return src.read().decode()
	except HTTPError as e:
		if e.status!=404:raise
		return None
def get_icons(package_name:str)->tuple[str,list[int]]:
	"""
	パッケージのアイコンを取得する。

	Params
	---
	package_name: アイコンを取得するパッケージの名前。

	Return
	---
	取得されたアイコンが保存されているフォルダのパスと、存在するアイコンの大きさのリストのタプル。
	アイコンの名前はそれぞれ、「x大きさ.png」
	"""
	escaped_name=escape_name(package_name)
	cache_path_=absjoin(CACHE_PATH,str(get_cache_id()))
	dst_path=absjoin(cache_path_,"meta/icons",escaped_name)
	_,repo_path=resolve_package_data(package_name)
	sizes=get_meta(package_name).icon_sizes
	if os.path.exists(dst_path):return (dst_path,sizes)
	os.makedirs(dst_path)
	base_url=urljoin(repo_path,"raw/iwashi_package/icons/")
	for size in sizes:
		file_name=f"x{size}.png"
		with download_and_open(urljoin(base_url,file_name)) as src:
			with open(absjoin(dst_path,file_name),"wb") as dst:
				dst.write(src.read())
	return (dst_path,sizes)

def get_screenshots(package_name:str)->tuple[str,int]:
	"""
	パッケージのスクショを取得する。

	Params
	---
	package_name: スクショを取得するパッケージの名前。

	Return
	---
	取得されたスクショが保存されているフォルダのパスと、スクショの数のタプル。
	スクショのファイルの名前はそれぞれ、「番号(1から始まる).png」
	"""
	escaped_name=escape_name(package_name)
	cache_path_=absjoin(CACHE_PATH,str(get_cache_id()))
	dst_path=absjoin(cache_path_,"meta/screenshots",escaped_name)
	_,repo_path=resolve_package_data(package_name)
	count=get_meta(package_name).screenshot_count
	if os.path.exists(dst_path):return (dst_path,count)
	os.makedirs(dst_path)
	base_url=urljoin(repo_path,"raw/iwashi_package/screenshots/")
	for index in range(count,0,-1):
		file_name=f"{index}.png"
		with download_and_open(urljoin(base_url,file_name)) as src:
			with open(absjoin(dst_path,file_name),"wb") as dst:
				dst.write(src.read())
	return (dst_path,count)

def save_index():
	index_={k:str(v)for k,v in index.items()}
	with open(PACKAGE_INDEX_PATH, "w") as fp:
		json.dump(index_,fp)
with open(PACKAGE_INDEX_PATH) as fp:
	index:dict[str,PackageInfo]={k:PackageInfo.read(v)for k,v in json.load(fp).items()}
	for v in index.values():
		if v.package_type==PackageType.auto:v.package_type=PackageType.bin
