import traceback
from .consts import ROOT_DIR as _ROOT_DIR
from .utils import PackageInfo as _PackageInfo
from abc import ABC as _ABC,abstractmethod as _abstractmethod
class IwashiError(Exception,_ABC):
	"""
	イワシが送出するエラーの基底クラス。
	"""
	@property
	@_abstractmethod
	def japanese_default_error_message(self)->str:
		"""
		日本語でのエラーメッセージ
		"""
		return traceback.format_exc()
class SelfUninstallationError(IwashiError):
	@property
	def japanese_default_error_message(self) -> str:
		return f"イワシはイワシ自身をアンインストールできません。ご自分で、「{_ROOT_DIR}」を削除してください。"
class PackageNotInstalledError(IwashiError):
	def __init__(self, package_name:str,*args: object, in_update:bool=False) -> None:
		super().__init__(*args)
		self.package_name=package_name
		self.in_update=in_update
	@property
	def japanese_default_error_message(self) -> str:
		res=f"パッケージ「{self.package_name}」がインストールされていません。"
		if self.in_update:res+="新たに追加する場合は、「install」を使ってください。"
		return res
class PackageAlreadyInstalledError(IwashiError):
	def __init__(self, package_name:str,*args: object, in_install:bool=False) -> None:
		super().__init__(*args)
		self.package_name=package_name
		self.in_install=in_install
	@property
	def japanese_default_error_message(self) -> str:
		res=f"パッケージ「{self.package_name}」はすでにインストールされています。"
		if self.in_install:res+="更新、もしくは、バージョンの変更をする場合は、「update」を使ってください。"
		return res
class UnknownPackageTypeError(IwashiError):
	def __init__(self, type:str,*args: object) -> None:
		super().__init__(*args)
		self.type=type
	@property
	def japanese_default_error_message(self) -> str:
		return f"「{self.type}」形式のパッケージに対応していません。"
class PackageInfoNotFoundError(IwashiError):
	def __init__(self, package_info:_PackageInfo,package_name:str,*args: object) -> None:
		super().__init__(*args)
		self.package_info=str(package_info)
		self.package_name=package_name
	@property
	def japanese_default_error_message(self) -> str:
		return f"パッケージ情報「{self.package_info}」をもつパッケージ「{self.package_name}」は存在しません。"
class InvaidPackageError(IwashiError):
	def __init__(self,package_name:str,*args: object) -> None:
		super().__init__(*args)
		self.package_name=package_name
	@property
	def japanese_default_error_message(self) -> str:
		return f"パッケージ「{self.package_name}」は不正です。"
class PackageNotFoundError(IwashiError):
	def __init__(self,package_name:str,*args: object) -> None:
		super().__init__(*args)
		self.package_name=package_name
	@property
	def japanese_default_error_message(self) -> str:
		return f"パッケージ「{self.package_name}」が見つかりませんでした。"
class InvaidVersionError(IwashiError,ValueError):
	def __init__(self,version:str,*args: object) -> None:
		super().__init__(*args)
		self.version=version
	@property
	def japanese_default_error_message(self) -> str:
		return f"バージョン「{self.version}」は不正です。詳しくは公式サイト(https://ubazame.duckdns.org/wiki/)をご覧ください。"
class InvaidPackageInfoError(IwashiError):
	def __init__(self, invaid_package_info:str,*args: object) -> None:
		super().__init__(*args)
		self.invaid_package_info=invaid_package_info
	@property
	def japanese_default_error_message(self) -> str:
		return f"パッケージ情報「{self.invaid_package_info}」は不正です。バージョンか、「@(パッケージの種類):(パッケージのバージョン)」のいずれかでなくてはいけません。"