import argparse
from functools import partial
import tqdm
from .imports import registries, save_config, uninstall, install, list_packages, list_versions, IwashiError, get_default_package_type, set_default_package_type, PackageType, PackageInfo, get_meta, get_readme, get_icons, get_screenshots, clear_cache, PackageMetaData, run
import sys

def command_uninstall(args):
	package_name:str=args.package_name
	if package_name=="katakuchi_iwashi":
		print("Error: カタクチイワシは、自分自身をアンインストールできません。")
		exit(-1)
	try:
		uninstall(package_name)
	except IwashiError as e:
		print("Error: "+e.japanese_default_error_message)
		exit(-1)
def command_install(args, is_update=False):
	td:tqdm.tqdm
	def start(size:int):
		nonlocal td;
		td=tqdm.tqdm(total=size,unit="B",unit_scale=True,file=sys.stderr)
	def step(step:int):
		td.update(step)
	def end():
		td.close()
	try:
		if not install(args.package_name,None if args.info is None else PackageInfo.read(args.info),is_update,progress_start=start,progress_step=step,progress_end=end):
			if args.info is None:
				print(f"パッケージ「{args.package_name}」は最新の状態です。")
			else:
				print(f"指定されたパッケージ情報「{args.info}」が現在のパッケージ情報と同一です。")
	except IwashiError as e:
		print("Error: "+e.japanese_default_error_message)
		exit(-1)
def command_list(args):
	print("利用可能なパッケージ一覧:"if args.registry else"現在のパッケージ一覧:")
	for name,value in list_packages(args.registry).items():
		print(f"{name}: {'(未インストール)'if value is None else value}")
def command_config(args):
	config:str=args.config
	value:str|None=args.value
	match config:
		case "default_package_type":
			if value is None:
				print(get_default_package_type().name)
			else:
				if value not in ["bin","py"]:
					print("Error: 「default_package_type」に設定できるのは、「bin」と「py」のどちらかです。")
					exit(-1)
				set_default_package_type(PackageType[value])
def command_meta(args):
	package_name:str=args.package_name
	type_:str|None=args.type
	meta_data:PackageMetaData|None=None
	if type_ in {None,"jp_name"}:
		if meta_data is None:meta_data=get_meta(package_name)
		print("パッケージの日本語名: "+meta_data.name['ja-jp'])
	if type_ in [None,"authors"]:
		if meta_data is None:meta_data=get_meta(package_name)
		print("パッケージの作者: "+("(なし)" if meta_data.authors is None else", ".join(meta_data.authors)))
	if type_ in [None,"readme"]:
		print("パッケージの説明:",end="")
		tmp_=get_readme(package_name)
		print("(なし)"if tmp_ is None else "\n"+tmp_)
	if type_ in [None,"icons"]:
		(tmp__,sizes)=get_icons(package_name)
		print(f"アイコン: 次の{len(sizes)}種類")
		[print(f"* 縦{s}px,横{s}px")for s in sizes]
		print(f"(すべて「{tmp__}」に保存)")
	if type_ in [None,"screenshots"]:
		(tmp___,number)=get_screenshots(package_name)
		print(f"スクリーンショット: {number}枚 (「{tmp___}」に保存)")
def command_run(args):
	package_name:str=args.package_name
	args_:list[str]=args.args
	try:
		run(package_name,*args_)
	except IwashiError as e:
		print("Error: "+e.japanese_default_error_message)
		exit(-1)
ap=argparse.ArgumentParser("katakuchi_iwashi", description="Umihebiのパッケージマネージャー")
ap.add_argument("-v","--version",action="version", version=list_packages()["katakuchi_iwashi"].version, help="これのバージョンを表示して、終了する。")
sub=ap.add_subparsers()
ap_install=sub.add_parser("install", help="パッケージをインストールする。")
ap_install.add_argument("package_name", help="パッケージ名")
ap_install.add_argument("info", help="パッケージのパッケージ情報", default=None, nargs="?")
ap_install.set_defaults(command=command_install)
ap_update=sub.add_parser("update", help="パッケージをアップデート(もしくは、指定したバージョンに変更)する。")
ap_update.add_argument("package_name", help="パッケージ名")
ap_update.add_argument("info", help="変更先のパッケージ情報", default=None, nargs="?")
ap_update.set_defaults(command=partial(command_install,is_update=True))
ap_uninstall=sub.add_parser("uninstall", help="パッケージをアンインストールする。")
ap_uninstall.add_argument("package_name", help="パッケージ名")
ap_uninstall.set_defaults(command=command_uninstall)
ap_list=sub.add_parser("list", help="パッケージとそのバージョンの一覧を表示する。")
ap_list.add_argument("-r","--registry", help="レジストリ上のパッケージを表示する。(インストールされていないパッケージのバージョンは「(未インストール)」と表示される。)",action="store_true")
ap_list.set_defaults(command=command_list)
ap_version=sub.add_parser("version", help="パッケージ、もしくは、マイワシのバージョンを表示する。")
ap_version.add_argument("package_name", help="パッケージ名(なければ、マイワシのバージョンを表示する)",nargs="?",default="maiwashi")
ap_version.add_argument("-a","--all", help="あれば、利用可能なバージョンをすべて表示する。",action="store_true")
ap_version.set_defaults(command=lambda args:print(", ".join(list_versions(args.package_name)) if args.all else list_packages().get(args.package_name,PackageInfo(f"Error: パッケージ「{args.package_name}」が存在しません。")).version))# ←汚いハック。
ap_version=sub.add_parser("package_info", help="パッケージのパッケージ情報を表示する。")
ap_version.add_argument("package_name", help="パッケージ名")
ap_version.set_defaults(command=lambda args:print(str(list_packages()[args.package_name])))
ap_registry=sub.add_parser("registry", help="レジストリを操作する。")
sub_registry=ap_registry.add_subparsers()
ap_registry_delete=sub_registry.add_parser("delete",help="レジストリを削除する。")
ap_registry_delete.add_argument("name", help="削除するレジストリの名前")
ap_registry_delete.set_defaults(command=lambda args:(registries.pop(args.name),save_config()))
ap_registry_get=sub_registry.add_parser("get",help="レジストリのURLを取得する。")
ap_registry_get.add_argument("name", help="取得するレジストリの名前")
ap_registry_get.set_defaults(command=lambda args:(print(f"{args.name}: {registries[args.name]}")))
ap_registry_set=sub_registry.add_parser("set",help="レジストリを変更、もしくは追加する。")
ap_registry_set.add_argument("name", help="変更するレジストリの名前(まだ存在しない場合は追加)")
ap_registry_set.add_argument("URL", help="レジストリのインデックスのURL")
ap_registry_set.set_defaults(command=lambda args:(registries.__setitem__(args.name,args.URL),save_config()))
ap_registry_list=sub_registry.add_parser("list",help="レジストリとそのURLの一覧を表示する。")
ap_registry_list.set_defaults(command=lambda _:[print(f"{name}: {url}")for name,url in registries.items()])
ap_config=sub.add_parser("config", help="設定を操作する。")
ap_config.add_argument("config",help="操作する設定。今は、既定のパッケージの種類、「default_package_type」のみ。",choices=["default_package_type"])
ap_config.add_argument("value",help="設定する値、なければ取得して表示する。",nargs="?",default=None)
ap_config.set_defaults(command=command_config)
ap_help=sub.add_parser("help", help="サブコマンド、もしくは、全体のヘルプを表示する。")
ap_help.add_argument("commands", help="サブコマンド(なければ、全体のヘルプ)",nargs="*")
ap_help.set_defaults(command=lambda args:ap.parse_args(args.commands+["-h"]))
ap_meta=sub.add_parser("meta", help="パッケージのメタデータを取得する。",description="typeに渡せるのは、日本語名を意味する「jp_name」、作者を意味する「authors」、説明を意味する「readme」、アイコンを意味する「icons」、スクリーンショットを意味する「screenshots」の内のどれか。")
ap_meta.add_argument("package_name", help="メタデータを取得するパッケージの名前")
ap_meta.add_argument("type", help="取得するメタデータの種類。渡せるものは上を参照。省略した場合はすべて。",default=None,nargs="?",choices=["jp_name","authors","readme","icons","screenshots"])
ap_meta.set_defaults(command=command_meta)
ap_clear_cache=sub.add_parser("clear_cache", help="キャッシュを破棄する。")
ap_clear_cache.set_defaults(command=lambda _:clear_cache())
ap_run=sub.add_parser("run",help="パッケージのエントリポイントを実行する。")
ap_run.add_argument("package_name",help="エントリポイントを実行するパッケージの名前")
ap_run.add_argument("args",nargs="*",help="エントリポイントに渡す引数。")
ap_clear_cache.set_defaults(command=lambda args:clear_cache())
def iwashi_main(argv:list[str]):
	args=ap.parse_args(argv[1:])
	if hasattr(args,"command"):args.command(args)
	else:ap.print_help()